/*
colormachinesdl.cc
Copyright 2007 Matt Heinzen

This file is part of Color Machine 0.8

Color Machine is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Color Machine is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Color Machine; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <time.h>
#include <stdio.h>

#include "colormachinesdl.h"
#include "utils.h"
#include "palettes.h"
#include "modules.h"


SDL_Surface *g_pScreen;
Uint8 *g_pScreenBuff;
SDL_Color g_Palette[256];
int g_DisplayWidth, g_DisplayHeight;


void MainLoop(int const ModuleCmd, int const PaletteCmd, int const EffectCmd,
	int const nModuleSecCmd, int const nCyclesCmd, int const nFPSCmd)
{
	SDL_Event event;
	int Module, Palette, Effect, PrevModule, PrevPalette, PrevEffect;
	int nModuleSec, nFrameTicks, nCycles;
	time_t StartTime;
	bool bReverse;
	clock_t StartClock;
	int i;

	// reset current and previous module and palette or set by parameter
	Module = ModuleCmd ? ModuleCmd : -1;
	Palette = PaletteCmd ? PaletteCmd : -1;
	PrevModule = -1;
	PrevPalette = -1;

	printf("ModuleCmd = %d, Module = %d\n", ModuleCmd, Module);

	// flush the event queue
	while (SDL_PollEvent(&event) == 1);

	// keep looping until a keypress, mouse click or quit event
	while (true)
	{
		// choose a new module and palette
		while (!ModuleCmd && Module == PrevModule)
			Module = Random(NUM_MODULES) + 1;
		while (!PaletteCmd && Palette == PrevPalette)
			Palette = Random(NUM_PALETTES) + 1;

		SDL_LockSurface(g_pScreen);

		// init the palette to all black
		InitPaletteBlack();

		// set the palette
		SDL_SetPalette(g_pScreen, SDL_PHYSPAL, g_Palette, 0, 256);

		// clear the screen
		memset(g_pScreenBuff, 0, g_DisplayWidth * g_DisplayHeight);

		// draw the chosen module
		switch (Module)
		{
		case 1: DrawModuleDiamonds(); break;
		case 2: DrawModuleGrid(); break;
		case 3: DrawModuleOrganic(); break;
		case 4: DrawModuleSpikes(); break;
		case 5: DrawModuleWorms(); break;
		case 6: DrawModulePipeline(); break;
		case 7: DrawModuleFractured(); break;
		case 8: DrawModuleCheckerboard(); break;
		case 9: DrawModuleMicrochip(); break;
		case 10: DrawModuleOpposition(); break;
		case 11: DrawModuleInfinity(); break;
		case 12: DrawModuleContour(); break;
		case 13: DrawModuleVortex(); break;
		case 14: DrawModuleThorns(); break;
		case 15: DrawModuleWaves(); break;
		case 16: DrawModuleNucleus(); break;
		case 17: DrawModuleJelly(); break;
		case 18: DrawModuleRays(); break;
		case 19: DrawModuleSpiral(); break;
		case 20: DrawModuleOsmosis(); break;
		case 21: DrawModuleMultipolarity(); break;
		case 22: DrawModuleRipples(); break;
		case 23: DrawModuleBeyond(); break;
		case 24: DrawModuleContinuum(); break;
		case 25: DrawModuleSpheres(); break;
		case 26: DrawModuleGodsOrb(); break;
		//case 27: DrawModuleOrion(); break;
		//case 28: DrawModule(); break;
		//case 29: DrawModule(); break;
		default: DrawModuleGodsOrb(); break;
		}

		// update the SDl surface
		SDL_UnlockSurface(g_pScreen);
		SDL_UpdateRect(g_pScreen, 0, 0, 0, 0);

		// initialize the chosen palette values
		switch (Palette)
		{
		case 1: InitPaletteRedPulse(); break;
		case 2: InitPaletteGreenPulse(); break;
		case 3: InitPaletteBluePulse(); break;
		case 4: InitPaletteYellowPulse(); break;
		case 5: InitPaletteCyanPulse(); break;
		case 6: InitPaletteMagentaPulse(); break;
		case 7: InitPaletteRainbow(); break;
		case 8: InitPaletteBrownian(); break;
		default: InitPaletteRainbow(); break;
		}

		// set the palette
		SDL_SetPalette(g_pScreen, SDL_PHYSPAL, g_Palette, 0, 256);

		// reset current and previous effect
		Effect = EffectCmd ? EffectCmd : -1;
		PrevEffect = -1;

		// generate semi-random lengths of time for the entire module and for
		// individual frames
		nModuleSec = nModuleSecCmd ? nModuleSecCmd
			: Random(MODULE_TIME_RANGE) + MIN_MODULE_TIME;
		nFrameTicks = nFPSCmd ? CLOCKS_PER_SEC / nFPSCmd
			: Random(CYCLE_TIME_RANGE) + MIN_CYCLE_TIME;

		// DEBUG
		//printf("nFrameTicks = %d\n", nFrameTicks);


		// start the timer for this module
		StartTime = time(NULL);

		// loop until the module timer has elapsed
		while (time(NULL) - StartTime < nModuleSec) {
			// select an effect
			while (!EffectCmd && Effect == PrevEffect)
				Effect = Random(NUM_EFFECTS) + 1;
			bReverse = Random(2) == 1;


			// TEST
			int R, G, B, RR, RG, RB;
			Effect = 0;
			do {
				R = Random(2);
				G = Random(2);
				B = Random(2);
				RR = Random(2);
				RG = Random(2);
				RB = Random(2);
			} while (R + G + B < 1);

			// TEST
			/**R = G = B = RR = RG = RB = 1;/**/


			// choose a random effect length
			nCycles = nCyclesCmd ? nCyclesCmd
				: Random(CYCLES_RANGE) + MIN_CYCLES;

			for (i = 0; i < nCycles; i++) {
				// start the timer for this frame
				StartClock = clock();

				// capture relevant events
				while (SDL_PollEvent(&event) == 1) {
					switch (event.type) {
					case SDL_KEYDOWN:
					case SDL_MOUSEBUTTONDOWN:
						printf("keyboard or mouse button\n");
						return;
					case SDL_QUIT:
						printf("quit\n");
						return;
					case SDL_MOUSEMOTION:
						break;
					}
				}

				// apply the current effect
				switch (Effect) {
				case 1 : RotatePalette(bReverse); break;
				case 2 : RotateRed(bReverse); break;
				case 3 : RotateGreen(bReverse); break;
				case 4 : RotateBlue(bReverse); break;
				}


				// TEST
				/**/if (R) RotateRed(RR);
				if (G) RotateGreen(RG);
				if (B) RotateBlue(RB);/**/

				// update SDL palette
				SDL_SetPalette(g_pScreen, SDL_PHYSPAL, g_Palette, 0, 256);


				// wait for the frame timer to elapse
				/**/while (clock() - StartClock < nFrameTicks) {}/**/


				// TEST: variable timing of cycles
				/**int wait;
				wait = int((cos(M_PI * 2.0 * i / nCycles) + 1.0) * nFrameTicks / 2 * 10);
				while (clock() - StartClock < wait) {}/**/
			}

			// store previous effect
			PrevEffect = Effect;
		}

		// fade to black
		for (i = 0; i < 64; i++)
		{
			StartClock = clock();
			FadePalette_x256(4);
			while (clock() - StartClock < CLOCKS_PER_SEC / 256) {}
		}

		// store previous module and palette
		PrevModule = Module;
		PrevPalette = Palette;
	}
}

int main(int argc, char **argv)
{

	// TODO: use command line arguments for selecting all parameters

	int Module, Palette, Effect, nModuleSec, nCycles, nFPS, RandSeed;
	char *pStr;
	time_t TT;
	int i;

	// initialize screensaver parameters from command line args
	Module = Palette = Effect = nModuleSec = nCycles = nFPS = RandSeed
		= 0;
	for (i = 1; i < argc; i++)
	{

		if ((pStr = strstr(argv[i], "m=")) != NULL)
			sscanf(pStr + 2, "%d", &Module);
		if ((pStr = strstr(argv[i], "p=")) != NULL)
			sscanf(pStr + 2, "%d", &Palette);
		if ((pStr = strstr(argv[i], "e=")) != NULL)
			sscanf(pStr + 2, "%d", &Effect);
		if ((pStr = strstr(argv[i], "t=")) != NULL)
			sscanf(pStr + 2, "%d", &nModuleSec);
		if ((pStr = strstr(argv[i], "c=")) != NULL)
			sscanf(pStr + 2, "%d", &nCycles);
		if ((pStr = strstr(argv[i], "f=")) != NULL)
			sscanf(pStr + 2, "%d", &nFPS);
		if ((pStr = strstr(argv[i], "r=")) != NULL)
			sscanf(pStr + 2, "%d", &RandSeed);
	}

	// initialze SDL
	if (SDL_Init(SDL_INIT_VIDEO) != 0)
	{
		printf("Unable to initialize SDL video\n");
		return 1;
	}

	// make sure SDL_Quit gets called when the program exits
	atexit(SDL_Quit);

	// attempt to set video mode with color index capabilities
	/**/g_pScreen = SDL_SetVideoMode(DEFAULT_WIDTH, DEFAULT_HEIGHT, 8,
		SDL_FULLSCREEN | SDL_HWPALETTE);/**/
	// DEBUG: for screenshots
	/**g_pScreen = SDL_SetVideoMode(640, 480, 8, SDL_HWPALETTE);/**/
	if (g_pScreen == NULL)
	{
		printf("Unable to set video mode: %s\n", SDL_GetError());
		return 1;
	}

	g_DisplayWidth = g_pScreen->w;
	g_DisplayHeight = g_pScreen->h;

	g_pScreenBuff = (Uint8 *)g_pScreen->pixels;

	// hide the mouse pointer
	SDL_ShowCursor(SDL_DISABLE);

	printf("display is %d by %d\n", g_DisplayWidth, g_DisplayHeight);

	// randomize with user selected seed
	if (RandSeed > 0) srandom(RandSeed);
	// randomize with the system clock
	else srandom(time(&TT));

	// start loop
	MainLoop(Module, Palette, Effect, nModuleSec, nCycles, nFPS);

	// DEBUGGING
	/*for (int i = 0; i < 256; i++)
		printf("%d: %d, %d, %d\n", i, g_Palette[i].r, g_Palette[i].g, g_Palette[i].b);*/

	return 0;
}
